import logging
import os
from glob import glob
from pathlib import Path
from typing import Iterator

from isdn.model import ISDNRecord, ISDNRoot
from pydantic import HttpUrl, ValidationError
from rdflib import URIRef

from ..namespace import *
from . import Literal, RDFMapper, _TripleMapType, bpo

logger = logging.getLogger(__name__)

DERIVATION_TYPE_MAP = {
    "オリジナル": ISDN.OriginalWork,
    "パロディ": ISDN.DerivativeWork,
}

PRODUCT_TYPE_MAP = {
    "同人誌": ISDN.Book,
    "同人グッズ": ISDN.DoujinGoods,
    "同人音楽": ISDN.Music,
    "同人ソフト": ISDN.Software,
    "同人ハード": ISDN.Hardware,
    "フィギュア・ドール": ISDN.FigurativeProduct,
    "衣装・アクセサリ": ISDN.Wearable,
    "フライヤー（ペーパー）": ISDN.Flyer,
    "イベントカタログ": ISDN.EventCatalog,
}

PRINTING_METHOD = {"オフセット印刷", "オンデマンド印刷", "コピー印刷"}
BINDING_METHOD = {"平綴じ", "中綴じ", "無線綴じ"}
CARRIER_TYPE = {"プレスCD", "プレスDVD", "CD-R", "DVD-R", "フロッピー"}


class ISDNXML2RDFMapper(RDFMapper):
    def __init__(self, xml_dir: str | os.PathLike):
        self.xml_dir = xml_dir

    def iterator(self) -> Iterator[ISDNRecord]:
        for file in glob(os.path.join(self.xml_dir, "*.xml")):
            yield ISDNRoot.from_xml_first(Path(file).read_bytes())

    @staticmethod
    def map_to_triples(row: ISDNRecord) -> tuple[URIRef, list[_TripleMapType]]:
        graph = None
        if row.rating_age == "15禁":
            graph = ISDN_GRAPH.ageRestricted15
        elif row.rating_age == "18禁":
            graph = ISDN_GRAPH.ageRestricted18

        s = ISDN_RES[row.isdn.code]

        triples = [
            (s, OWL.sameAs, ISDN_RES_PURL[row.isdn.code]),
            (s, RDF.type, ISDN.DoujinProduct),
            (s, RDF.type, PRODUCT_TYPE_MAP[row.type]),
            (s, ISDN.isdn, row.isdn.code),
            (
                s,
                SCHEMA.identifier,
                bpo(
                    [
                        (RDF.type, SCHEMA.PropertyValue),
                        (RDF.type, ISDN.ISDNIdentifier),
                        (SCHEMA.propertyID, "ISDN"),
                        (SCHEMA.value, Literal(row.isdn.code)),
                        (SCHEMA.alternateName, row.disp_isdn),
                        (ISDN.isdnPrefix, row.isdn.prefix),
                        (ISDN.isdnRegistrationGroup, row.isdn.group),
                        (ISDN.isdnRegistrationGroupName, row.region),
                        (ISDN.isdnRegistrant, row.isdn.registrant),
                        (ISDN.isdnPublication, row.isdn.publication),
                        (ISDN.isdnCheckDigit, row.isdn.check_digit),
                    ]
                ),
            ),
            (s, ISDN.derivationType, DERIVATION_TYPE_MAP[row.class_]),
            (s, SCHEMA.category, row.type),
            (s, ISDN.ratingGender, row.rating_gender),
            (s, ISDN.ratingAge, row.rating_age),
            (s, SCHEMA.name, row.product_name),
            (s, RDFS.label, Literal(row.product_name, lang="ja")),
            (s, RDFS.label, Literal(row.product_yomi, lang="ja-Hira")),
            (
                s,
                SCHEMA.publisher,
                bpo(
                    [
                        (RDF.type, SCHEMA.Organization),
                        (SCHEMA.name, row.publisher_name),
                        (RDFS.label, Literal(row.publisher_name, lang="ja")),
                        (RDFS.label, Literal(row.publisher_yomi, lang="ja-Hira")),
                        (ISDN.isdnRegistrant, row.publisher_code),
                    ]
                ),
            ),
            (s, SCHEMA.datePublished, Literal(row.issue_date, datatype=XSD.date)),
            (
                s,
                SCHEMA.genre,
                bpo(
                    [
                        (RDF.type, ISDN.ComiketGenre),
                        (SCHEMA.name, row.genre_name),
                        (SCHEMA.identifier, row.genre_code),
                        (SCHEMA.description, row.genre_user),
                    ]
                ),
            ),
            (
                s,
                SCHEMA.genre,
                bpo(
                    [
                        (RDF.type, ISDN.CCode),
                        (SCHEMA.identifier, row.c_code),
                        (ISDN.cCodeTargetAudience, row.author),
                        (ISDN.cCodePublicationForm, row.shape),
                        (ISDN.cCodeContentCategory, row.contents),
                    ]
                ),
            ),
            (
                s,
                SCHEMA.offers,
                bpo(
                    [
                        (RDF.type, SCHEMA.Offer),
                        (SCHEMA.price, Literal(row.price, datatype=XSD.decimal)),
                        (SCHEMA.priceCurrency, row.price_unit),
                    ]
                ),
            ),
            (s, ISDN.bookJanCodeSecondRow, row.barcode2),
            (s, SCHEMA.description, row.product_comment),
            (s, ISDN.productStyle, row.product_style),
            (s, SCHEMA.size, row.product_size),
            (
                s,
                ISDN.carrierExtent,
                bpo(
                    [
                        (RDF.type, SCHEMA.QuantitativeValue),
                        (SCHEMA.value, Literal(row.product_capacity, datatype=XSD.decimal)),
                        (SCHEMA.unitText, row.product_capacity_unit),
                    ]
                ),
            ),
            (s, SCHEMA.thumbnailUrl, URIRef(str(row.sample_image_uri)) if row.sample_image_uri else None),
        ]

        if row.product_style:
            for part in row.product_style.split("・"):
                if part in PRINTING_METHOD:
                    triples.append((s, ISDN.printingMethod, part))
                elif part in BINDING_METHOD:
                    triples.append((s, ISDN.bindingMethod, part))
                elif part in CARRIER_TYPE:
                    triples.append((s, ISDN.carrierType, part))

        if row.product_capacity_unit == "ページ":
            triples.append((s, SCHEMA.numberOfPages, Literal(row.product_capacity, datatype=XSD.decimal)))

        for option in row.useroptions:
            triples.append(
                (
                    s,
                    SCHEMA.additionalProperty,
                    bpo(
                        [
                            (RDF.type, SCHEMA.PropertyValue),
                            (RDF.type, ISDN.UserOption),
                            (SCHEMA.name, option.property),
                            (SCHEMA.value, option.value),
                        ]
                    ),
                )
            )

        for link in row.external_links:
            try:
                link_url = HttpUrl(link.uri.strip())
            except ValidationError:
                logger.warning(f"Invalid external link: {link.uri} (ISDN: {row.isdn.code})")
                continue
            triples.append(
                (
                    s,
                    SCHEMA.relatedLink,
                    bpo(
                        [
                            (RDF.type, SCHEMA.WebPage),
                            (SCHEMA.name, link.title),
                            (SCHEMA.url, URIRef(str(link_url))),
                        ]
                    ),
                )
            )

        return graph, triples


__all__ = ["ISDNXML2RDFMapper"]



from os import PathLike

import click
from isal import igzip_threaded
from joblib.parallel import cpu_count

from . import __version__
from .mappers import RDFFormatType
from .mappers.isdn import ISDNXML2RDFMapper


@click.group()
def cli():
    pass


@cli.command()
def version():
    click.echo(f"isdn_ld/{__version__}")


@cli.command(help="Convert XML files to RDF")
@click.argument("xml_dir", type=click.Path(exists=True, file_okay=False))
@click.argument("output_file", type=click.Path(writable=True))
@click.option(
    "--processes",
    "-p",
    type=int,
    default=max(1, cpu_count(only_physical_cores=True) - 1),
    help="Number of worker processes",
)
@click.option(
    "--io-threads",
    type=int,
    default=2,
    help="This is only valid if the '--compress' option is specified.",
)
@click.option(
    "--format",
    "-f",
    "_format",
    type=click.Choice([v.name for v in RDFFormatType]),
    default=RDFFormatType.nq.name,
)
@click.option("--compress", "-c", is_flag=True, help="Enable gzip compression")
def convert(
    xml_dir: PathLike,
    output_file: PathLike,
    processes: int,
    io_threads: int,
    _format: str,
    compress: bool,
):
    mapper = ISDNXML2RDFMapper(xml_dir)

    click.echo(f"output: {output_file}")
    click.echo("Running mapper ...")

    f = igzip_threaded.open(output_file, "wt", threads=io_threads) if compress else open(output_file, "w")
    mapper.run(n_jobs=processes, output=f, format_type=RDFFormatType[_format])
    f.close()



import sys
from abc import ABC, abstractmethod
from enum import Enum
from typing import IO, Any, Iterator, Tuple, Union

from joblib import Parallel, delayed
from rdflib import BNode, URIRef
from rdflib import Literal as LiteralRdflib
from rdflib.graph import _ObjectType, _PredicateType, _SubjectType, _TripleType
from rdflib.namespace import RDF
from rdflib.plugins.serializers.nquads import _nq_row
from rdflib.plugins.serializers.nt import _nt_row

_TripleMapType = Tuple[_SubjectType, _PredicateType, Union[str, _ObjectType, "BlankPredicateObjectMap"]]
_PredicateObjectType = Tuple[_PredicateType, Union[str, _ObjectType, "BlankPredicateObjectMap"]]


class RDFFormatType(Enum):
    nt = "nt"
    nq = "nq"


class Literal:
    def __new__(cls, value, *args, **kwargs) -> LiteralRdflib | None:
        if value is None:
            return None
        return LiteralRdflib(value, *args, **kwargs)


class BlankPredicateObjectMap:
    def __init__(self, predicate_objects: list[_PredicateObjectType]):
        self.subject = BNode()
        self.predicate_objects = predicate_objects


def bpo(predicate_objects: list[_PredicateObjectType]) -> BlankPredicateObjectMap:
    return BlankPredicateObjectMap(predicate_objects)


def normalize_triple(triple: _TripleType) -> _TripleMapType | None:
    if triple[2] is None:
        return None
    if isinstance(triple[2], Literal) and triple[2] is None:
        return None
    if type(triple[2]) is str:
        triple = (triple[0], triple[1], Literal(triple[2]))
    return triple


def flatten_triple_map(triple_map: _TripleMapType) -> list[_TripleType]:
    triples = []

    if isinstance(triple_map[2], BlankPredicateObjectMap):
        bnode = triple_map[2].subject
        b_triples = []
        for po in triple_map[2].predicate_objects:
            b_triples.extend(flatten_triple_map((bnode, po[0], po[1])))

        # Ignore statements if rdf:type only
        if len([t for t in b_triples if t[1] != RDF.type]) > 0:
            triples.append((triple_map[0], triple_map[1], bnode))
            triples.extend(b_triples)
    else:
        t = normalize_triple(triple_map)
        if t:
            triples.append(t)

    return triples


def serialize_triple(triple: _TripleMapType, graph: URIRef | None = None) -> str:
    return _nq_row(triple, graph) if graph else _nt_row(triple)


class RDFMapper(ABC):
    @abstractmethod
    def iterator(self) -> Iterator:
        pass

    def run(
        self,
        n_jobs: int = -1,
        output: IO[str] = sys.stdout,
        format_type: RDFFormatType = RDFFormatType.nq,
    ):
        def job(row: Any) -> str:
            graph_uri, triple_maps = self.map_to_triples(row)
            return "".join(
                [
                    serialize_triple(
                        triple,
                        graph=graph_uri if format_type == RDFFormatType.nq else None,
                    )
                    for triple_map in triple_maps
                    for triple in flatten_triple_map(triple_map)
                ]
            )

        res = Parallel(n_jobs=n_jobs, return_as="generator_unordered", verbose=1)(
            delayed(job)(row) for row in self.iterator()
        )
        for lines in res:
            output.write(lines)

    @staticmethod
    @abstractmethod
    def map_to_triples(row: Any) -> tuple[URIRef, list[_TripleMapType]]:
        """
        :param row: Object provided by the iterator
        :return: ({Graph URI}, [(subject, predicate, object) ...])
        """
        raise NotImplementedError


__all__ = ["RDFFormatType", "RDFMapper"]



from rdflib import Namespace
from rdflib.namespace import OWL, RDF, RDFS, XSD, ClosedNamespace

SCHEMA = Namespace("http://schema.org/")
ISDN_RES = Namespace("http://metadata.moe/isdn/res/")
ISDN_RES_PURL = Namespace("http://purl.org/isdn/")

ISDN = ClosedNamespace(
    "http://metadata.moe/ns/isdn/",
    [
        # Classes
        "ISDNIdentifier",
        "DoujinProduct",
        "ComiketGenre",
        "CCode",
        "UserOption",
        ## 形態
        "Book",
        "DoujinGoods",
        "Music",
        "Software",
        "Hardware",
        "FigurativeProduct",
        "Wearable",
        "Flyer",
        "EventCatalog",
        ## 大分類
        "OriginalWork",
        "DerivativeWork",
        # Properties
        "isdn",
        "isdnPrefix",
        "isdnRegistrationGroup",
        "isdnRegistrationGroupName",
        "isdnRegistrant",
        "isdnPublication",
        "isdnCheckDigit",
        "derivationType",
        "ratingGender",
        "ratingAge",
        "cCodeTargetAudience",
        "cCodePublicationForm",
        "cCodeContentCategory",
        "bookJanCodeSecondRow",
        "productStyle",
        "printingMethod",
        "bindingMethod",
        "carrierType",
        "carrierExtent",
    ],
)

ISDN_GRAPH = ClosedNamespace("http://metadata.moe/isdn/graph/", ["ageRestricted15", "ageRestricted18"])


__all__ = [
    "OWL",
    "RDF",
    "RDFS",
    "XSD",
    "SCHEMA",
    "ISDN",
    "ISDN_RES",
    "ISDN_RES_PURL",
    "ISDN_GRAPH",
]
